import 'dart:convert';
import 'dart:io';

import 'package:basic_utils/basic_utils.dart';
import 'package:cli_a2soft/cli/cli.dart';
import 'package:cli_a2soft/cli/cli_mixin.dart';
import 'package:args/command_runner.dart';
import 'package:cli_a2soft/commands/swagger_command/body_model_type.dart';
import 'package:cli_a2soft/commands/swagger_command/operation.dart';
import 'package:conduit_open_api_fork/v3.dart';
import 'package:http/http.dart' as http;

part 'freezed_model_template.dart';

class SwaggerCommand extends Command with CLIMixin {
  SwaggerCommand();

  @override
  String get description => "Создает freezed-модели по ссылке на swagger";

  @override
  String get name => "swagger";

  String generateQueryModel({
    required String filename,
    required String filenameCamelCase,
    required String nestedModelName,
    required List<APIParameter?> parameters,
    String? comment,
  }) {
    String parametersString = "";
    for (var element in parameters) {
      if (element != null) {
        parametersString += generateParameter(
          description: element.description ?? "",
          type: element.schema?.type ?? APIType.string,
          name: element.name ?? "",
        );
      }
    }
    return freezedTemplate(
      filename: filename,
      filenameCamelCase: filenameCamelCase,
      nestedModelName: nestedModelName,
      comment: comment,
      parameters: parametersString,
    );
  }

  Future<String> generateBodyModel({
    required APISchemaObject rootSchema,
    required String directoryPath,
    required Operation operation,
    required BodyModelType modelType,
    String? prefixName,
  }) async {
    final fileName =
        "${operation.prefixSnakeCase}${prefixName != null ? "_${StringUtils.camelCaseToLowerUnderscore(prefixName)}" : ""}_${modelType.lower}";
    final fileNameCamelCase =
        "${operation.prefixPascalCase}${prefixName ?? ""}${modelType.capital}";
    final rootFile = await File("$directoryPath/$fileName.dart").create();
    String parametersString = '';
    String importsString = '';
    if (rootSchema.properties != null) {
      for (final key in rootSchema.properties!.keys) {
        final value = rootSchema.properties![key];
        switch (value?.type) {
          case APIType.array:
            final modelName = await generateBodyModel(
              rootSchema: value!.items!,
              directoryPath: directoryPath,
              operation: operation,
              modelType: modelType,
              prefixName: StringUtils.capitalize(key) + "Item",
            );
            importsString += generateImport(StringUtils.camelCaseToLowerUnderscore(modelName));
            parametersString += generateParameter(
              description: value.description ?? "",
              modelType: "List<$modelName>",
              name: key,
            );
            break;
          case APIType.object:
            final modelName = await generateBodyModel(
              rootSchema: value?.items ?? APISchemaObject(),
              directoryPath: directoryPath,
              operation: operation,
              modelType: modelType,
              prefixName: StringUtils.capitalize(key),
            );
            importsString += generateImport(StringUtils.camelCaseToLowerUnderscore(modelName));
            parametersString += generateParameter(
              description: value?.description ?? "",
              modelType: modelName,
              name: key,
            );
            break;
          default:
            parametersString += generateParameter(
              description: value?.description ?? "",
              type: value?.type,
              name: key,
            );
        }
      }
    }
    final content = freezedTemplate(
      filename: fileName,
      filenameCamelCase: fileNameCamelCase,
      nestedModelName: prefixName ?? operation.prefixPascalCase + modelType.nestedModel,
      comment: rootSchema.description,
      parameters: parametersString,
      imports: importsString,
    );
    await rootFile.writeAsString(content);
    await CLI.runFormatter(rootFile.parent.path, printErrors: false);
    return fileNameCamelCase;
  }

  String _getSnakeCaseFolderPath(String path, String prefix) => path
      .replaceAll(prefix, "")
      .split('/')
      .map((e) => e.replaceAll('{', '').replaceAll('}', ''))
      .map((element) => StringUtils.camelCaseToLowerUnderscore(element))
      .join('/');

  @override
  Future<void> run() async {
    /// Запрашиваем ссылку на swagger.json
    final swaggerUrl = getUserInput<String>(
      helpPhrase: "Введите ссылку на swagger док в json-формате",
      defaultValue: "",
    );

    final json = await http.get(Uri.parse(swaggerUrl), headers: {"Accept": "application/json"});

    final Map<String, dynamic> jsonMap = jsonDecode(json.body);

    /// Сваггер-док в структуре дарта
    final doc = APIDocument.fromMap(jsonMap);

    /// Префикс к методам в сваггере
    final prefix = getUserInput<String>(
      helpPhrase: "Префикс к методам в сваггере",
      defaultValue: "",
    );

    /// Запрашиваем путь для папки, куда будут генерироваться модели
    final currentDirectory = "models";
    final pathToFolder = getUserInput<String>(
      helpPhrase: "Введите путь для генерации моделей",
      defaultValue: currentDirectory,
    );

    /// Словарь с папкой и методами из этой папки
    final Map<String, List<Operation>> foldersMap = {};

    doc.paths?.forEach(
      (key, value) {
        /// Название метода без префикса
        final methodWithoutPrefix = _getSnakeCaseFolderPath(key, prefix);

        /// Название папки
        final folderName = methodWithoutPrefix.split('/').first;

        /// Название метода
        final methodName = methodWithoutPrefix.split('/').join('_');

        // Фильтруем на null
        final entries = value?.operations.entries
                .where((element) => element.value != null)
                .map((e) => MapEntry(e.key, e.value as APIOperation))
                .toList() ??
            [];

        final operations = entries
            .map((e) => Operation(
                  operation: e.value,
                  folderName: methodName,
                  methodType: e.key,
                ))
            .toList();

        if (foldersMap.containsKey(folderName)) {
          foldersMap[folderName]!.addAll(operations);
        } else {
          foldersMap.putIfAbsent(folderName, () => operations);
        }
      },
    );

    for (final folder in foldersMap.entries) {
      final rootFolder = await Directory('$pathToFolder/${folder.key}').create(recursive: true);

      for (final operation in folder.value) {
        final operationFolder =
            await Directory('${rootFolder.path}/${operation.prefixSnakeCase}').create();
        final apiOperation = operation.operation;

        // Обработка модели реквеста (query)
        if (apiOperation.parameters?.isNotEmpty ?? false) {
          final params = apiOperation.parameters!;
          final fileName = "${operation.prefixSnakeCase}_request";
          final fileNameCamelCase = "${operation.prefixPascalCase}Request";
          final nestedModelName = "${operation.prefixPascalCase}Req";
          final file = await File("${operationFolder.path}/$fileName.dart").create();
          await file.writeAsString(
            generateQueryModel(
              filename: fileName,
              filenameCamelCase: fileNameCamelCase,
              nestedModelName: nestedModelName,
              parameters: params,
              comment: apiOperation.summary,
            ),
          );
        }

        // Обработка модели реквеста (body)
        if (apiOperation.requestBody?.content?.containsKey("application/json") ?? false) {
          final rootSchema = apiOperation.requestBody!.content!["application/json"]!.schema;
          if (rootSchema != null) {
            await generateBodyModel(
              rootSchema: rootSchema,
              directoryPath: operationFolder.path,
              modelType: BodyModelType(ModelType.request),
              operation: operation,
            );
          }
        }

        // Обработка модели респонса
        if (apiOperation.responses?.containsKey('200') ?? false) {
          final response = apiOperation.responses!['200']!;
          if (response.content?.containsKey("application/json") ?? false) {
            final rootSchema = response.content!["application/json"]!.schema;
            if (rootSchema != null) {
              await generateBodyModel(
                rootSchema: rootSchema,
                directoryPath: operationFolder.path,
                modelType: BodyModelType(ModelType.response),
                operation: operation,
              );
            }
          }
        }
      }
    }
  }
}
